@@ -14,7 +14,7 @@
 
 """SVGPath <=> skia-pathops constructs to enable ops on paths."""
 import functools
-import pathops
+import pathops  # pytype: disable=import-error
 from typing import Sequence, Tuple
 from picosvg.svg_meta import SVGCommand, SVGCommandGen, SVGCommandSeq
 from picosvg.svg_transform import Affine2D
