@@ -14,8 +14,18 @@
 
 import copy
 import dataclasses
+from itertools import zip_longest
 from picosvg.geometric_types import Point, Rect
-from picosvg import svg_meta
+from picosvg.svg_meta import (
+    check_cmd,
+    cmd_coords,
+    number_or_percentage,
+    parse_css_declarations,
+    path_segment,
+    strip_ns,
+    SVGCommand,
+    SVGCommandSeq,
+)
 from picosvg import svg_pathops
 from picosvg.arc_to_cubic import arc_to_cubic
 from picosvg.svg_path_iter import parse_svg_path
@@ -23,6 +33,10 @@ from picosvg.svg_transform import Affine2D
 from typing import Generator, Iterable
 
 
+def _round_multiple(f: float, of: float) -> float:
+    return round(f / of) * of
+
+
 def _explicit_lines_callback(subpath_start, curr_pos, cmd, args, *_):
     del subpath_start
     if cmd == "v":
@@ -44,22 +58,36 @@ def _explicit_lines_callback(subpath_start, curr_pos, cmd, args, *_):
     return ((cmd, args),)
 
 
-def _relative_to_absolute(curr_pos, cmd, args):
-    x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
-    if cmd.islower():
-        cmd = cmd.upper()
+def _rewrite_coords(cmd_converter, coord_converter, curr_pos, cmd, args):
+    x_coord_idxs, y_coord_idxs = cmd_coords(cmd)
+    desired_cmd = cmd_converter(cmd)
+    if cmd != desired_cmd:
+        cmd = desired_cmd
+        # if x_coord_idxs or y_coord_idxs:
         args = list(args)  # we'd like to mutate 'em
         for x_coord_idx in x_coord_idxs:
-            args[x_coord_idx] += curr_pos.x
+            args[x_coord_idx] += coord_converter(curr_pos.x)
         for y_coord_idx in y_coord_idxs:
-            args[y_coord_idx] += curr_pos.y
+            args[y_coord_idx] += coord_converter(curr_pos.y)
 
     return (cmd, tuple(args))
 
 
-def _next_pos(curr_pos, cmd, cmd_args):
+def _relative_to_absolute(curr_pos, cmd, args):
+    return _rewrite_coords(
+        lambda cmd: cmd.upper(), lambda curr_scaler: curr_scaler, curr_pos, cmd, args
+    )
+
+
+def _absolute_to_relative(curr_pos, cmd, args):
+    return _rewrite_coords(
+        lambda cmd: cmd.lower(), lambda curr_scaler: -curr_scaler, curr_pos, cmd, args
+    )
+
+
+def _next_pos(curr_pos, cmd, cmd_args) -> Point:
     # update current position
-    x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
+    x_coord_idxs, y_coord_idxs = cmd_coords(cmd)
     new_x, new_y = curr_pos
     if cmd.isupper():
         if x_coord_idxs:
@@ -79,17 +107,18 @@ def _move_endpoint(curr_pos, cmd, cmd_args, new_endpoint):
     # we need to be able to alter both axes
     ((cmd, cmd_args),) = _explicit_lines_callback(None, curr_pos, cmd, cmd_args)
 
-    x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
-    cmd_args = list(cmd_args)  # we'd like to mutate
-    new_x, new_y = new_endpoint
-    if cmd.islower():
-        new_x = new_x - curr_pos.x
-        new_y = new_y - curr_pos.y
+    x_coord_idxs, y_coord_idxs = cmd_coords(cmd)
+    if x_coord_idxs or y_coord_idxs:
+        cmd_args = list(cmd_args)  # we'd like to mutate
+        new_x, new_y = new_endpoint
+        if cmd.islower():
+            new_x = new_x - curr_pos.x
+            new_y = new_y - curr_pos.y
 
-    cmd_args[x_coord_idxs[-1]] = new_x
-    cmd_args[y_coord_idxs[-1]] = new_y
+        cmd_args[x_coord_idxs[-1]] = new_x
+        cmd_args[y_coord_idxs[-1]] = new_y
 
-    return cmd, cmd_args
+    return cmd, tuple(cmd_args)
 
 
 # Subset of https://www.w3.org/TR/SVG11/painting.html
@@ -111,6 +140,7 @@ class SVGShape:
     opacity: float = 1.0
     transform: str = ""
     style: str = ""
+    display: str = "inline"
 
     def _copy_common_fields(
         self,
@@ -130,6 +160,7 @@ class SVGShape:
         opacity,
         transform,
         style,
+        display,
     ):
         self.id = id
         self.clip_path = clip_path
@@ -147,10 +178,16 @@ class SVGShape:
         self.opacity = opacity
         self.transform = transform
         self.style = style
+        self.display = display
 
     def might_paint(self) -> bool:
         """False if we're sure this shape will not paint. True if it *might* paint."""
 
+        shape = self.apply_style_attribute()
+
+        if shape.display == "none":
+            return False
+
         def _visible(fill, opacity):
             return fill != "none" and opacity != 0
             # we're ignoring fill-opacity
@@ -159,8 +196,8 @@ class SVGShape:
         if all(c[0].upper() == "M" for c in self.as_cmd_seq()):
             return False
 
-        return _visible(self.stroke, self.stroke_opacity) or _visible(
-            self.fill, self.opacity
+        return _visible(shape.stroke, shape.stroke_opacity) or _visible(
+            shape.fill, shape.opacity
         )
 
     def bounding_box(self) -> Rect:
@@ -179,7 +216,7 @@ class SVGShape:
     def as_path(self) -> "SVGPath":
         raise NotImplementedError("You should implement as_path")
 
-    def as_cmd_seq(self) -> svg_meta.SVGCommandSeq:
+    def as_cmd_seq(self) -> SVGCommandSeq:
         return (
             self.as_path()
             .explicit_lines()  # hHvV => lL
@@ -193,7 +230,7 @@ class SVGShape:
         # only meaningful for path, which overrides
         return self
 
-    def stroke_commands(self, tolerance) -> Generator[svg_meta.SVGCommand, None, None]:
+    def stroke_commands(self, tolerance) -> Generator[SVGCommand, None, None]:
         return svg_pathops.stroke(
             self.as_cmd_seq(),
             self.stroke_linecap,
@@ -217,7 +254,7 @@ class SVGShape:
                 f.name.replace("_", "-"): f.type for f in dataclasses.fields(self)
             }
             raw_attrs = {}
-            unparsed_style = svg_meta.parse_css_declarations(
+            unparsed_style = parse_css_declarations(
                 target.style, raw_attrs, property_names=attr_types.keys()
             )
             for attr_name, attr_value in raw_attrs.items():
@@ -238,14 +275,31 @@ class SVGShape:
                 setattr(target, field.name, round(field_value, ndigits))
         return target
 
-    def almost_equals(self, other: "SVGShape", tolerance: int) -> bool:
-        assert isinstance(other, SVGShape)
-        return self.round_floats(tolerance) == other.round_floats(tolerance)
+    def round_multiple(self, multiple_of: float, inplace=False) -> "SVGShape":
+        """Round all floats in SVGShape to nearest multiple of multiple_of."""
+        target = self
+        if not inplace:
+            target = copy.deepcopy(self)
+        for field in dataclasses.fields(target):
+            field_value = getattr(self, field.name)
+            if isinstance(field_value, float):
+                setattr(target, field.name, _round_multiple(field_value, multiple_of))
+        return target
+
+    def almost_equals(self, other: "SVGShape", tolerance: float) -> bool:
+        for (l_cmd, l_args), (r_cmd, r_args) in zip_longest(
+            self.as_path(), other.as_path(), fillvalue=(None, ())
+        ):
+            if l_cmd != r_cmd or len(l_args) != len(r_args):
+                return False
+            if any(abs(lv - rv) > tolerance for lv, rv in zip(l_args, r_args)):
+                return False
+        return True
 
 
 # https://www.w3.org/TR/SVG11/paths.html#PathElement
 @dataclasses.dataclass
-class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
+class SVGPath(SVGShape, SVGCommandSeq):
     d: str = ""
 
     def __init__(self, **kwargs):
@@ -258,7 +312,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
         self.d += path_snippet
 
     def _add_cmd(self, cmd, *args):
-        self._add(svg_meta.path_segment(cmd, *args))
+        self._add(path_segment(cmd, *args))
 
     def M(self, *args):
         self._add_cmd("M", *args)
@@ -267,7 +321,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
         self._add_cmd("m", *args)
 
     def _arc(self, c, rx, ry, x, y, large_arc):
-        self._add(svg_meta.path_segment(c, rx, ry, 0, large_arc, 1, x, y))
+        self._add(path_segment(c, rx, ry, 0, large_arc, 1, x, y))
 
     def A(self, rx, ry, x, y, large_arc=0):
         self._arc("A", rx, ry, x, y, large_arc)
@@ -300,7 +354,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
         self._add_cmd("Q", *args)
 
     def end(self):
-        self._add("z")
+        self._add("Z")
 
     def as_path(self) -> "SVGPath":
         return self
@@ -317,7 +371,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
     def __iter__(self):
         return parse_svg_path(self.d, exploded=True)
 
-    def walk(self, callback):
+    def walk(self, callback) -> "SVGPath":
         """Walk path and call callback to build potentially new commands.
 
         https://www.w3.org/TR/SVG11/paths.html
@@ -332,7 +386,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
 
         # iteration gives us exploded commands
         for idx, (cmd, args) in enumerate(self):
-            svg_meta.check_cmd(cmd, args)
+            check_cmd(cmd, args)
             if idx == 0 and cmd == "m":
                 cmd = "M"
 
@@ -355,6 +409,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
         self.d = ""
         for _, cmd, args in new_cmds:
             self._add_cmd(cmd, *args)
+        return self
 
     def move(self, dx, dy, inplace=False):
         """Returns a new path that is this one shifted."""
@@ -366,7 +421,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
             # Shift the absolute parts and call it a day.
             if cmd.islower():
                 return ((cmd, args),)
-            x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
+            x_coord_idxs, y_coord_idxs = cmd_coords(cmd)
             args = list(args)  # we'd like to mutate 'em
             for x_coord_idx in x_coord_idxs:
                 args[x_coord_idx] += dx
@@ -380,11 +435,9 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
         target.walk(move_callback)
         return target
 
-    def absolute(self, inplace=False) -> "SVGPath":
-        """Returns equivalent path with only absolute commands."""
-
-        def absolute_callback(subpath_start, curr_pos, cmd, args, *_):
-            new_cmd, new_cmd_args = _relative_to_absolute(curr_pos, cmd, args)
+    def _rewrite_path(self, rewrite_fn, inplace) -> "SVGPath":
+        def rewrite_callback(subpath_start, curr_pos, cmd, args, *_):
+            new_cmd, new_cmd_args = rewrite_fn(curr_pos, cmd, args)
 
             # if we modified cmd to pass *very* close to subpath start snap to it
             # eliminates issues with not-quite-closed shapes due float imprecision
@@ -398,9 +451,17 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
         target = self
         if not inplace:
             target = copy.deepcopy(self)
-        target.walk(absolute_callback)
+        target.walk(rewrite_callback)
         return target
 
+    def absolute(self, inplace=False) -> "SVGPath":
+        """Returns equivalent path with only absolute commands."""
+        return self._rewrite_path(_relative_to_absolute, inplace)
+
+    def relative(self, inplace=False) -> "SVGPath":
+        """Returns equivalent path with only relative commands."""
+        return self._rewrite_path(_absolute_to_relative, inplace)
+
     def explicit_lines(self, inplace=False):
         """Replace all vertical/horizontal lines with line to (x,y)."""
         target = self
@@ -485,13 +546,11 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
         return target
 
     @classmethod
-    def from_commands(
-        cls, svg_cmds: Generator[svg_meta.SVGCommand, None, None]
-    ) -> "SVGPath":
+    def from_commands(cls, svg_cmds: Generator[SVGCommand, None, None]) -> "SVGPath":
         return cls().update_path(svg_cmds, inplace=True)
 
     def update_path(
-        self, svg_cmds: Generator[svg_meta.SVGCommand, None, None], inplace=False
+        self, svg_cmds: Generator[SVGCommand, None, None], inplace=False
     ) -> "SVGPath":
         target = self
         if not inplace:
@@ -507,7 +566,7 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
 
         Also reformat the SVGPath.d string floats with the same rounding.
         """
-        target: SVGPath = super().round_floats(ndigits, inplace=inplace)
+        target: SVGPath = super().round_floats(ndigits, inplace=inplace).as_path()
 
         d, target.d = target.d, ""
         for cmd, args in parse_svg_path(d):
@@ -515,6 +574,19 @@ class SVGPath(SVGShape, svg_meta.SVGCommandSeq):
 
         return target
 
+    def round_multiple(self, multiple_of: float, inplace=False) -> "SVGPath":
+        """Round all floats in SVGPath to given decimal digits.
+
+        Also reformat the SVGPath.d string floats with the same rounding.
+        """
+        target: SVGPath = super().round_multiple(multiple_of, inplace=inplace).as_path()
+
+        d, target.d = target.d, ""
+        for cmd, args in parse_svg_path(d):
+            target._add_cmd(cmd, *(_round_multiple(n, multiple_of) for n in args))
+
+        return target
+
 
 # https://www.w3.org/TR/SVG11/shapes.html#CircleElement
 @dataclasses.dataclass
@@ -575,7 +647,7 @@ class SVGPolygon(SVGShape):
     def as_path(self) -> SVGPath:
         *shape_fields, points = dataclasses.astuple(self)
         if self.points:
-            path = SVGPath(d="M" + self.points + " z")
+            path = SVGPath(d="M" + self.points + " Z")
         else:
             path = SVGPath()
         path._copy_common_fields(*shape_fields)
@@ -637,17 +709,97 @@ class SVGRect(SVGShape):
         return path
 
 
-def union(shapes: Iterable[SVGShape]) -> SVGPath:
-    return SVGPath.from_commands(
-        svg_pathops.union(
-            [s.as_cmd_seq() for s in shapes], [s.clip_rule for s in shapes]
+def _get_gradient_units_relative_scale(grad_el, view_box):
+    gradient_units = grad_el.attrib.get("gradientUnits", "objectBoundingBox")
+    if gradient_units == "userSpaceOnUse":
+        # For gradientUnits="userSpaceOnUse", percentages represent values relative to
+        # the current viewport.
+        return view_box
+    elif gradient_units == "objectBoundingBox":
+        # For gradientUnits="objectBoundingBox", percentages represent values relative
+        # to the object bounding box. The latter defines an abstract coordinate system
+        # with origin at (0,0) and a nominal width and height = 1.
+        return Rect(0, 0, 1, 1)
+    else:
+        raise ValueError(
+            f'{strip_ns(grad_el.tag)} gradientUnits="{gradient_units}" not supported'
         )
+
+
+def _parse_common_gradient_parts(gradient, el, view_box):
+    self = gradient
+    self.gradientUnits = _get_gradient_units_relative_scale(el, view_box)
+    if "gradientTransform" in el.attrib:
+        self.gradientTransform = Affine2D.fromstring(el.attrib["gradientTransform"])
+    if "spreadMethod" in el.attrib:
+        self.spreadMethod = el.attrib["spreadMethod"]
+    return self.gradientUnits.w, self.gradientUnits.h
+
+
+# https://developer.mozilla.org/en-US/docs/Web/SVG/Element/linearGradient
+# Should be parsed with from_element
+@dataclasses.dataclass
+class SVGLinearGradient:
+    x1: float = 0.0
+    x2: float = 0.0
+    y1: float = 0.0
+    y2: float = 0.0
+    gradientUnits: Rect = Rect(0, 0, 1, 1)
+    gradientTransform: Affine2D = Affine2D.identity()
+    spreadMethod: str = "pad"
+
+    @staticmethod
+    def from_element(el, view_box) -> "SVGLinearGradient":
+        self = SVGLinearGradient()
+        width, height = _parse_common_gradient_parts(self, el, view_box)
+
+        self.x1 = number_or_percentage(el.attrib.get("x1", "0%"), width)
+        self.y1 = number_or_percentage(el.attrib.get("y1", "0%"), height)
+        self.x2 = number_or_percentage(el.attrib.get("x2", "100%"), width)
+        self.y2 = number_or_percentage(el.attrib.get("y2", "0%"), height)
+        return self
+
+
+# https://developer.mozilla.org/en-US/docs/Web/SVG/Element/radialGradient
+# Should be parsed with from_element
+@dataclasses.dataclass
+class SVGRadialGradient:
+    cx: float = 0.0
+    cy: float = 0.0
+    r: float = 0.0
+    fr: float = 0.0
+    fx: float = 0.0
+    fy: float = 0.0
+    gradientUnits: Rect = Rect(0, 0, 1, 1)
+    gradientTransform: Affine2D = Affine2D.identity()
+    spreadMethod: str = "pad"
+
+    @staticmethod
+    def from_element(el, view_box) -> "SVGRadialGradient":
+        self = SVGRadialGradient()
+        width, height = _parse_common_gradient_parts(self, el, view_box)
+
+        self.cx = number_or_percentage(el.attrib.get("cx", "50%"), width)
+        self.cy = number_or_percentage(el.attrib.get("cy", "50%"), height)
+        self.r = number_or_percentage(el.attrib.get("r", "50%"), width)
+
+        raw_fx = el.attrib.get("fx")
+        self.fx = number_or_percentage(raw_fx, width) if raw_fx is not None else self.cx
+        raw_fy = el.attrib.get("fy")
+        self.fy = (
+            number_or_percentage(raw_fy, height) if raw_fy is not None else self.cy
+        )
+        self.fr = number_or_percentage(el.attrib.get("fr", "0%"), width)
+        return self
+
+
+def union(shapes: Iterable[SVGShape]) -> Generator[SVGCommand, None, None]:
+    return svg_pathops.union(
+        [s.as_cmd_seq() for s in shapes], [s.clip_rule for s in shapes]
     )
 
 
-def intersection(shapes: Iterable[SVGShape]) -> SVGPath:
-    return SVGPath.from_commands(
-        svg_pathops.intersection(
-            [s.as_cmd_seq() for s in shapes], [s.clip_rule for s in shapes]
-        )
+def intersection(shapes: Iterable[SVGShape]) -> Generator[SVGCommand, None, None]:
+    return svg_pathops.intersection(
+        [s.as_cmd_seq() for s in shapes], [s.clip_rule for s in shapes]
     )
